"""
Sequence data object layer.

It's not clear that it's really a ``Sequence`` that we want, in the sense of
``collections.abc`` (https://docs.python.org/3/library/collections.abc.html).
Namely, we don't care about ``Reversible`` in most cases.
Perhaps we need an intermediate type?

"""

from collections.abc import Sequence


# No wrapper hooks have been added (yet)
# This was just to verify that ``Sequence`` mixins can resolve all methods from
# just __len__ and __getitem__ (but how, I don't know yet, seems it's burried in C)
class Seq(Sequence):
    def __init__(self, seq):
        self.seq = seq

    def __len__(self):
        return len(self.seq)

    def __getitem__(self, item):
        return self.seq[item]



"""main creek objects"""
from creek.base import Creek

from creek.infinite_sequence import InfiniteSeq, IndexedBuffer
from creek.tools import (
    filter_and_index_stream,
    dynamically_index,
    DynamicIndexer,
    count_increments,
    size_increments,
    BufferStats,
    Segmenter,
)



"""Utils for creek"""

from functools import (
    WRAPPER_ASSIGNMENTS,
    partial,
    update_wrapper as _update_wrapper,
    wraps as _wraps,
)
from itertools import islice

from typing import Any, Iterable, Iterator, Union, NewType
from typing import Protocol, runtime_checkable

from functools import singledispatch, partial


def iterize(func, name=None):
    """From an In->Out function, makes a Iterator[In]->Itertor[Out] function.

    >>> f = lambda x: x * 10
    >>> f(2)
    20
    >>> iterized_f = iterize(f)
    >>> list(iterized_f(iter([1,2,3])))
    [10, 20, 30]

    """
    iterized_func = partial(map, func)
    if name is not None:
        iterized_func.__name__ = name
    return iterized_func


IteratorItem = Any


@runtime_checkable
class IterableType(Protocol):
    """An iterable type that can actually be used in singledispatch

    >>> assert isinstance([1, 2, 3], IterableType)
    >>> assert not isinstance(2, IterableType)
    """

    def __iter__(self) -> Iterable[IteratorItem]:
        pass


@runtime_checkable
class IteratorType(Protocol):
    """An iterator type that can actually be used in singledispatch

    >>> assert isinstance(iter([1, 2, 3]), IteratorType)
    >>> assert not isinstance([1, 2, 3], IteratorType)
    """

    def __next__(self) -> IteratorItem:
        pass


@runtime_checkable
class CursorFunc(Protocol):
    """An argument-less function returning an iterator's values"""

    def __call__(self) -> IteratorItem:
        """Get the next iterator's item and increment the cursor"""


IterType = NewType('IterType', Union[IteratorType, IterableType, CursorFunc])
IterType.__doc__ = 'A type that can be made into an iterator'

wrapper_assignments = (*WRAPPER_ASSIGNMENTS, '__defaults__', '__kwdefaults__')
update_wrapper = partial(_update_wrapper, assigned=wrapper_assignments)
wraps = partial(_wraps, assigned=wrapper_assignments)


# ---------------------------------------------------------------------------------------
# iteratable, iterator, cursors
# TODO: If bring i2 as dependency, use mk_sentinel here
no_sentinel = type('no_sentinel', (), {})()
no_default = type('no_default', (), {})()


class IteratorExit(BaseException):
    """Raised when an iterator should quit being iterated on, signaling this event
    any process that cares to catch the signal.
    We chose to inherit directly from `BaseException` instead of `Exception`
    for the same reason that `GeneratorExit` does: Because it's not technically
    an error.

    See: https://docs.python.org/3/library/exceptions.html#GeneratorExit
    """


DFLT_INTERRUPT_EXCEPTIONS = (StopIteration, IteratorExit, KeyboardInterrupt)


def iterate_until_exception(iterator, interrupt_exceptions=DFLT_INTERRUPT_EXCEPTIONS):
    while True:
        try:
            next(iterator)
        except interrupt_exceptions:
            print('ending')
            break


def iterable_to_iterator(iterable: Iterable, sentinel=no_sentinel) -> Iterator:
    """Get an iterator from an iterable

    >>> iterable = [1, 2, 3]
    >>> iterator = iterable_to_iterator(iterable)
    >>> assert isinstance(iterator, Iterator)
    >>> assert list(iterator) == iterable

    You can also specify a sentinel, which will result in the iterator stoping just
    before it encounters that sentinel value

    >>> iterable = [1, 2, 3, 4, None, None, 7]
    >>> iterator = iterable_to_iterator(iterable, None)
    >>> assert isinstance(iterator, Iterator)
    >>> list(iterator)
    [1, 2, 3, 4]
    """
    if sentinel is no_sentinel:
        return iter(iterable)
    else:
        return iter(iter(iterable).__next__, sentinel)


def iterator_to_cursor(iterator: Iterator, default=no_default) -> CursorFunc:
    """Get a cursor function for the input iterator.

    >>> iterator = iter([1, 2, 3])
    >>> cursor = iterator_to_cursor(iterator)
    >>> assert callable(cursor)
    >>> assert cursor() == 1
    >>> assert list(cursor_to_iterator(cursor)) == [2, 3]

    Note how we consumed the cursor till the end; by using cursor_to_iterator.
    Indeed, `list(iter(cursor))` wouldn't have worked since a cursor isn't a iterator,
    but a callable to get the items an the iterator would give you.

    You can specify a default. The default has the same role that it has for the
    `next` function: It makes the cursor function return that default when the iterator
    has been "consumed" (i.e. would raise a `StopIteration`).

    >>> iterator = iter([1, 2])
    >>> cursor = iterator_to_cursor(iterator, None)
    >>> assert callable(cursor)
    >>> cursor()
    1
    >>> cursor()
    2

    And then...

    >>> assert cursor() is None
    >>> assert cursor() is None

    forever.

    """
    if default is no_default:
        return partial(next, iterator)
    else:
        return partial(next, iterator, default)


def cursor_to_iterator(cursor: CursorFunc, sentinel=no_sentinel) -> Iterator:
    """Get an iterator from a cursor function.

    A cursor function is a callable that you call (without arguments) to get items of
    data one by one.

    Sometimes, especially in live io contexts, that's the kind interface you're given
    to consume a stream.

    >>> cursor = iter([1, 2, 3]).__next__
    >>> assert not isinstance(cursor, Iterator)
    >>> assert not isinstance(cursor, Iterable)
    >>> assert callable(cursor)

    If you want to consume your stream as an iterator instead, use `cursor_to_iterator`.

    >>> iterator = cursor_to_iterator(cursor)
    >>> assert isinstance(iterator, Iterator)
    >>> list(iterator)
    [1, 2, 3]

    If you want your iterator to stop (without a fuss) as soon as the cursor returns a
    particular element (called a sentinel), say it:

    >>> cursor = iter([1, 2, None, None, 3]).__next__
    >>> iterator = cursor_to_iterator(cursor, sentinel=None)
    >>> list(iterator)
    [1, 2]

    """
    return iter(cursor, sentinel)


def iterable_to_cursor(iterable: Iterable) -> CursorFunc:
    """Get a cursor function from an iterable.
    """
    iterator = iterable_to_iterator(iterable)
    return iterator_to_cursor(iterator)


@singledispatch
def to_iterator(x: IteratorType, sentinel=no_sentinel):
    """Get an iterator from an iterable or a cursor function

    >>> from typing import Iterator
    >>> it = to_iterator([1, 2, 3])
    >>> assert isinstance(it, Iterator)
    >>> list(it)
    [1, 2, 3]
    >>> list(it)
    []

    >>> cursor = iter([1, 2, 3]).__next__
    >>> assert isinstance(cursor, CursorFunc)
    >>> it = to_iterator(cursor)
    >>> assert isinstance(it, Iterator)
    >>> list(it)
    [1, 2, 3]
    >>> list(it)
    []

    You can use sentinels too

    >>> list(to_iterator([1, 2, None, 4], sentinel=None))
    [1, 2]
    >>> cursor = iter([1, 2, 3, 4, 5]).__next__
    >>> list(to_iterator(cursor, sentinel=4))
    [1, 2, 3]
    """
    if sentinel is no_sentinel:
        return x
    else:
        cursor = x.__next__
        return iter(cursor, sentinel)


@to_iterator.register
def _(x: IterableType, sentinel=no_sentinel):
    return to_iterator.__wrapped__(iter(x), sentinel)
    # TODO: Use of __wrapped__ seems hacky. Better way?
    # TODO: Why does to_iterator(iter(x), sentinel) lead to infinite recursion?


@to_iterator.register
def _(x: CursorFunc, sentinel=no_sentinel):
    return iter(x, sentinel)


# ---------------------------------------------------------------------------------------

no_such_item = type('NoSuchItem', (), {})()


class stream_util:
    def always_true(*args, **kwargs):
        return True

    def do_nothing(*args, **kwargs):
        pass

    def rewind(self, instance):
        instance.seek(0)

    def skip_lines(self, instance, n_lines_to_skip=0):
        instance.seek(0)


class PreIter:
    def skip_items(self, instance, n):
        return islice(instance, n, None)


def cls_wrap(cls, obj):
    if isinstance(obj, type):

        @wraps(obj, updated=())
        class Wrap(cls):
            @wraps(obj.__init__)
            def __init__(self, *args, **kwargs):
                wrapped = obj(*args, **kwargs)
                super().__init__(wrapped)

        # Wrap.__signature__ = signature(obj)

        return Wrap
    else:
        return cls(obj)


# TODO: Make identity_func "identifiable". If we use the following one, we can use == to detect it's use,
# TODO: ... but there may be a way to annotate, register, or type any identity function so it can be detected.


def identity_func(x):
    return x


static_identity_method = staticmethod(identity_func)


from inspect import signature, Signature


# Note: Pipe code is completely independent (with inspect imports signature & Signature)
#  If you only need simple pipelines, use this, or even copy/paste it where needed.
# TODO: Public interface mis-aligned with i2. funcs list here, in i2 it's dict. Align?
#  If we do so, it would be a breaking change since any dependents that expect funcs
#  to be a list of funcs will iterate over a iterable of names instead.
class Pipe:
    """Simple function composition. That is, gives you a callable that implements input -> f_1 -> ... -> f_n -> output.

    >>> def foo(a, b=2):
    ...     return a + b
    >>> f = Pipe(foo, lambda x: print(f"x: {x}"))
    >>> f(3)
    x: 5
    >>> len(f)
    2

    You can name functions, but this would just be for documentation purposes.
    The names are completely ignored.

    >>> g = Pipe(
    ...     add_numbers = lambda x, y: x + y,
    ...     multiply_by_2 = lambda x: x * 2,
    ...     stringify = str
    ... )
    >>> g(2, 3)
    '10'
    >>> len(g)
    3

    Notes:
        - Pipe instances don't have a __name__ etc. So some expectations of normal functions are not met.
        - Pipe instance are pickalable (as long as the functions that compose them are)

    You can specify a single functions:

    >>> Pipe(lambda x: x + 1)(2)
    3

    but

    >>> Pipe()
    Traceback (most recent call last):
      ...
    ValueError: You need to specify at least one function!

    You can specify an instance name and/or doc with the special (reserved) argument
    names ``__name__`` and ``__doc__`` (which therefore can't be used as function names):

    >>> f = Pipe(map, add_it=sum, __name__='map_and_sum', __doc__='Apply func and add')
    >>> f(lambda x: x * 10, [1, 2, 3])
    60
    >>> f.__name__
    'map_and_sum'
    >>> f.__doc__
    'Apply func and add'

    """

    funcs = ()

    def __init__(self, *funcs, **named_funcs):
        named_funcs = self._process_reserved_names(named_funcs)
        funcs = list(funcs) + list(named_funcs.values())
        self.funcs = funcs
        n_funcs = len(funcs)
        if n_funcs == 0:
            raise ValueError('You need to specify at least one function!')

        elif n_funcs == 1:
            other_funcs = ()
            first_func = last_func = funcs[0]
        else:
            first_func, *other_funcs = funcs
            *_, last_func = other_funcs

        self.__signature__ = Pipe._signature_from_first_and_last_func(
            first_func, last_func
        )
        self.first_func, self.other_funcs = first_func, other_funcs

    _reserved_names = ('__name__', '__doc__')

    def _process_reserved_names(self, named_funcs):
        for name in self._reserved_names:
            if (value := named_funcs.pop(name, None)) is not None:
                setattr(self, name, value)
        return named_funcs

    def __call__(self, *args, **kwargs):
        out = self.first_func(*args, **kwargs)
        for func in self.other_funcs:
            out = func(out)
        return out

    def __len__(self):
        return len(self.funcs)

    _dflt_signature = Signature.from_callable(lambda *args, **kwargs: None)

    @staticmethod
    def _signature_from_first_and_last_func(first_func, last_func):
        try:
            input_params = signature(first_func).parameters.values()
        except ValueError:  # function doesn't have a signature, so take default
            input_params = Pipe._dflt_signature.parameters.values()
        try:
            return_annotation = signature(last_func).return_annotation
        except ValueError:  # function doesn't have a signature, so take default
            return_annotation = Pipe._dflt_signature.return_annotation
        return Signature(tuple(input_params), return_annotation=return_annotation)



from creek.multi_streams import *

from warnings import warn

warn(f'Moved to creek.multi_streams')

from dataclasses import dataclass

from typing import Any, NewType, Callable, Iterable, Union, Tuple
from numbers import Number


def MyType(
    name: str,
    tp: type = Any,
    doc: Optional[str] = None,
    aka: Optional[Union[str, Iterable[str]]] = None,
):
    """Like `typing.NewType` with some extras (`__doc__` and `_aka` attributes, etc.)
    """

    new_tp = NewType(name, tp)
    if doc is not None:
        new_tp.__doc__ = doc
    if aka is not None:
        new_tp._aka = aka
    return new_tp


TimeIndex = MyType(
    'TimeIndex',
    Number,
    doc='A number indexing time. Could be in an actual time unit, or could just be '
    'an enumerator (i.e. "ordinal time")',
)
BT = MyType(
    'BT',
    TimeIndex,
    doc='TimeIndex for the lower bound of an interval of time. '
    'Stands for "Bottom Time". By convention, a BT is inclusive.',
)
TT = MyType(
    'TT',
    TimeIndex,
    doc='TimeIndex for the upper bound of an interval of time. '
    'Stands for "Upper Time". By convention, a TT is exclusive.',
)
IntervalTuple = MyType(
    'IntervalTuple',
    Tuple[BT, TT],
    doc='Denotes an interval of time by specifying the (BT, TT) pair',
)
IntervalSlice = MyType(
    'IntervalSlice',
    slice,  # Note: extra condition: non-None .start and .end, and no .step
    doc='Denotes an interval of time by specifying the (BT, TT) pair',
)

Intervals = Iterable[IntervalTuple]
BufferedIntervals = Intervals
RetrievedIntervals = Intervals
QueryInterval = Intervals
IntervalsRetriever = Callable[[QueryInterval, BufferedIntervals], RetrievedIntervals]


@dataclass
class IntervalSlicer:
    match_intervals: IntervalsRetriever



"""Tools for multi-streams"""

from itertools import product
from typing import Mapping, Iterable, Any, Optional, Callable
import heapq
from dataclasses import dataclass
from functools import partial
from operator import itemgetter

from creek.util import Pipe, identity_func

StreamsMap = Mapping[Any, Iterable]  # a map of {name: stream} pairs


@dataclass
class MergedStreams:
    """Creates an iterable of (stream_id, stream_item) pairs from a stream Mapping,
    that is, {stream_id: stream, ...}.

    The stream_item will be yield in sorted order.
    Sort behavior can be modified by the ``sort_key`` argument which behaves like ``key``
    arguments of built-in like ``sorted``, ``heapq.merge``, ``itertools.groupby``, etc.

    If given, the `sort_key` function applies to ``stream_item`` (not to ``stream_id``).

    Important: To function as expected, the streams should be already sorted (according
    to the ``sort_key`` order).

    The cannonical use case of this function is to "flatten", or "weave together"
    multiple streams of timestamped data. We're given several streams that provide
    ``(timestamp, data)`` items (where timestamps arrive in order within each stream)
    and we get a single stream of ``(stream_id, (timestamp, data))`` items where
    the ``timestamp``s are yield in sorted order.

    The following example uses a dict pointing to a fixed-size list as the ``stream_map``
    but in general the ``stream_map`` will be a ``Mapping`` (not necessarily a dict)
    whose values are potentially bound-less streams.

    >>> streams_map = {
    ...     'hello': [(2, 'two'), (3, 'three'), (5, 'five')],
    ...     'world': [(0, 'zero'), (1, 'one'), (3, 'three'), (6, 'six')]
    ... }
    >>> streams_items = MergedStreams(streams_map)
    >>> it = iter(streams_items)
    >>> list(it)  # doctest: +NORMALIZE_WHITESPACE
    [('world', (0, 'zero')),
     ('world', (1, 'one')),
     ('hello', (2, 'two')),
     ('hello', (3, 'three')),
     ('world', (3, 'three')),
     ('hello', (5, 'five')),
     ('world', (6, 'six'))]
    """

    streams_map: StreamsMap
    sort_key: Optional[Callable] = None

    def __post_init__(self):
        if self.sort_key is None:
            self.effective_sort_key = itemgetter(1)
        else:
            self.effective_sort_key = Pipe(itemgetter(1), self.sort_key)

    def __iter__(self):
        for item in heapq.merge(
            *multi_stream_items(self.streams_map), key=self.effective_sort_key
        ):
            yield item


def multi_stream_items(streams_map: StreamsMap):
    """Provides a iterable of (k1, v1_1), (k1, v1_2), ...
    
    >>> streams_map = {'hello': 'abc', 'world': [1, 2]}
    >>> hello_items, world_items = multi_stream_items(streams_map)
    >>> list(hello_items)
    [('hello', 'a'), ('hello', 'b'), ('hello', 'c')]
    >>> list(world_items)
    [('world', 1), ('world', 2)]
    """
    for stream_id, stream in streams_map.items():
        yield product([stream_id], stream)


def transform_methods(cls, method_trans=staticmethod):
    """Applies method_trans to all the methods of `cls`

    >>> from functools import partial
    >>> staticmethods = partial(transform_methods, method_trans=staticmethod)

    Now staticmethods is a class decorator that can be used to make all methods
    be defined as staticmethods in bulk

    >>> @staticmethods
    ... class C:
    ...     foo = lambda x: x + 1
    ...     bar = lambda y: y * 2
    >>> c = C()
    >>> c.foo(6)
    7
    >>> c.bar(6)
    12
    """
    for attr_name in vars(cls):
        attr = getattr(cls, attr_name)
        if callable(attr):
            setattr(cls, attr_name, method_trans(attr))

    return cls


staticmethods = partial(transform_methods, method_trans=staticmethod)


@staticmethods
class SortKeys:
    all_but_last = itemgetter(-1)
    second_item = itemgetter(1)



"""Tools to work with creek objects"""

import time
from collections import deque
from typing import (
    Any,
    Callable,
    Tuple,
    TypeVar,
    Callable,
    Any,
    Iterable,
    Sequence,
    cast,
)
from dataclasses import dataclass
from itertools import chain
from operator import itemgetter
from functools import partial

from creek.util import Pipe

Index = Any
DataItem = TypeVar('DataItem')
# TODO: Could have more args. How to specify this in typing?
IndexUpdater = Callable[[Index, DataItem], Index]
Indexer = Callable[[DataItem], Tuple[Index, DataItem]]

from itertools import chain


# TODO: Possible performance enhancement by using class with precompiled slices
# TODO: Compare with apply_and_fanout (no cast here, and slice instead of :)
# Note: I hesitated to make the signature (func, apply_to_idx, seq) instead
# Note: This would have allowed to use partial(apply_func_to_index, func, idx)
# Note: instead, but NOT partial(apply_func_to_index, func=func, apply_to_idx=idx)!!
def apply_func_to_index(seq, apply_to_idx, func):
    """
    >>> apply_func_to_index([1, 2, 3], 1, lambda x: x * 10)
    (1, 20, 3)

    If you're going to apply the same function to the same index, you might
    want to partialize ``apply_func_to_index`` to be able to reuse it simply:

    >>> from functools import partial
    >>> f = partial(apply_func_to_index, apply_to_idx=0, func=str.upper)
    >>> list(map(f, ['abc', 'defgh']))
    [('A', 'b', 'c'), ('D', 'e', 'f', 'g', 'h')]

    """
    apply_to_element, *_ = seq[slice(apply_to_idx, apply_to_idx + 1)]
    return tuple(
        chain(
            seq[slice(None, apply_to_idx)],
            [func(apply_to_element)],
            seq[slice(apply_to_idx + 1, None)],
        )
    )


def apply_and_fanout(
    seq: Sequence, func: Callable[[Any], Iterable], idx: int
) -> Iterable[tuple]:
    """Apply function (that returns an Iterable) to an element of a sequence
    and fanout (broadcast) the resulting items, to produce an iterable of tuples each
    containing
    one of these items along with 'a copy' of the other tuple elements

    >>> list(apply_and_fanout([1, 'abc', 3], iter, 1))
    [(1, 'a', 3), (1, 'b', 3), (1, 'c', 3)]
    >>> list(apply_and_fanout(['bob', 'alice', 2], lambda x: x * ['hi'], 2))
    [('bob', 'alice', 'hi'), ('bob', 'alice', 'hi')]
    >>> list(apply_and_fanout(["bob", "alice", 2], lambda x: x.upper(), 1))
    [('bob', 'A', 2), ('bob', 'L', 2), ('bob', 'I', 2), ('bob', 'C', 2), ('bob', 'E', 2)]

    See Also:
        ``fanout_and_flatten`` and ``fanout_and_flatten_dicts``
    """
    seq = tuple(seq)  # TODO: Overhead: Should we impose seq to be tuple?
    # TODO: See how apply_func_to_index takes care of this problem with chain
    left_seq = seq[0 : max(idx, 0)]  # TODO: Use seq[None:idx] instead?
    right_seq = seq[(idx + 1) :]
    return (left_seq + (item,) + right_seq for item in func(seq[idx]))


def fanout_and_flatten(iterable_of_seqs, func, idx, aggregator=chain.from_iterable):
    """Apply apply_and_fanout to an iterable of sequences.

    >>> seq_iterable = [('abcdef', 'first'), ('ghij', 'second')]
    >>> func = lambda a: zip(*([iter(a)] * 2))  # func is a chunker
    >>> assert list(fanout_and_flatten(seq_iterable, func, 0)) == [
    ...     (('a', 'b'), 'first'),
    ...     (('c', 'd'), 'first'),
    ...     (('e', 'f'), 'first'),
    ...     (('g', 'h'), 'second'),
    ...     (('i', 'j'), 'second')
    ... ]
    """
    apply = partial(apply_and_fanout, func=func, idx=idx)
    return aggregator(map(apply, iterable_of_seqs))


def fanout_and_flatten_dicts(
    iterable_of_dicts, func, fields, idx_field, aggregator=chain.from_iterable
):
    """Apply apply_and_fanout to an iterable of dicts.

    >>> iterable_of_dicts = [
    ...     {'wf': 'abcdef', 'tag': 'first'}, {'wf': 'ghij', 'tag': 'second'}
    ... ]
    >>> func = lambda a: zip(*([iter(a)] * 2))  # func is a chunker
    >>> fields = ['wf', 'tag']
    >>> idx_field = 'wf'
    >>> assert list(
    ...     fanout_and_flatten_dicts(iterable_of_dicts, func, fields, idx_field)) == [
    ...         {'wf': ('a', 'b'), 'tag': 'first'},
    ...         {'wf': ('c', 'd'), 'tag': 'first'},
    ...         {'wf': ('e', 'f'), 'tag': 'first'},
    ...         {'wf': ('g', 'h'), 'tag': 'second'},
    ...         {'wf': ('i', 'j'), 'tag': 'second'}
    ... ]
    """
    egress = Pipe(partial(zip, fields), dict)
    return map(
        egress,
        fanout_and_flatten(
            map(itemgetter(*fields), iterable_of_dicts),
            func=func,
            idx=fields.index(idx_field),
            aggregator=aggregator,
        ),
    )


def filter_and_index_stream(
    stream: Iterable, data_item_filt, timestamper: Indexer = enumerate
):
    """Index a stream and filter it (based only on the data items).

    >>> assert (
    ... list(filter_and_index_stream('this  is   a   stream', data_item_filt=' ')) == [
    ... (0, 't'),
    ... (1, 'h'),
    ... (2, 'i'),
    ... (3, 's'),
    ... (6, 'i'),
    ... (7, 's'),
    ... (11, 'a'),
    ... (15, 's'),
    ... (16, 't'),
    ... (17, 'r'),
    ... (18, 'e'),
    ... (19, 'a'),
    ... (20, 'm')
    ... ])

    >>> list(filter_and_index_stream(
    ...     [1, 2, 3, 4, 5, 6, 7, 8],
    ...     data_item_filt=lambda x: x % 2))
    [(0, 1), (2, 3), (4, 5), (6, 7)]
    """
    if not callable(data_item_filt):
        sentinel = data_item_filt
        data_item_filt = lambda x: x != sentinel
    return filter(lambda x: data_item_filt(x[1]), timestamper(stream))


# TODO: Refactor dynamic indexing set up so that
#  IndexUpdater = Callable[[DataItem, Index, ...], Index] (data and index inversed)
#  Rationale: One can (and usually would want to) have a default current_idx, which
#  can be used as the start index too.
count_increments: IndexUpdater


def count_increments(current_idx: Index, obj: DataItem, step=1):
    return current_idx + step


size_increments: IndexUpdater


def size_increments(current_idx, obj: DataItem, size_func=len):
    return current_idx + size_func(obj)


current_time: IndexUpdater


def current_time(current_idx, obj):
    """Doesn't even look at current_idx or obj. Just gives the current time"""
    return time.time()


@dataclass
class DynamicIndexer:
    """
    :param start: The index to start at (the first data item will have this index)
    :param idx_updater: The (Index, DataItem) -> Index

    Let's take a finite stream of finite iterables (strings here):

    >>> stream = ['stream', 'of', 'different', 'sized', 'chunks']

    The default ``DynamicIndexer`` just does what ``enumerate`` does:

    >>> counter_index = DynamicIndexer()
    >>> list(map(counter_index, stream))
    [(0, 'stream'), (1, 'of'), (2, 'different'), (3, 'sized'), (4, 'chunks')]

    That's because it uses the default ``idx_updater`` function just increments by one.
    This function, DynamicIndexer.count_increments, is shown below

    >>> def count_increments(current_idx, data_item, step=1):
    ...     return current_idx + step

    To get the index starting at 10, we can specify ``start=10``, and to step the
    index by 3 we can partialize ``count_increments``:

    >>> from functools import partial
    >>> step3 = partial(count_increments, step=3)
    >>> list(map(DynamicIndexer(start=10, idx_updater=step3), stream))
    [(10, 'stream'), (13, 'of'), (16, 'different'), (19, 'sized'), (22, 'chunks')]

    You can specify any custom ``idx_updater`` you want: The requirements being that
    this function should take ``(current_idx, data_item)`` as the input, and
    return the next "current index", that is, what the index of the next data item will
    be.
    Note that ``count_increments`` ignored the ``data_item`` completely, but sometimes
    you want to take the data item into account.
    For example, your data item may contain several elements, and you want your
    index to index these elements, therefore you should update your index by
    incrementing it with the number of elements.

    We have ``DynamicIndexer.size_increments`` for that, the code is shown below:

    >>> def size_increments(current_idx, data_item, size_func=len):
    ...     return current_idx + size_func(data_item)
    >>> size_index = DynamicIndexer(idx_updater=DynamicIndexer.size_increments)
    >>> list(map(size_index, stream))
    [(0, 'stream'), (6, 'of'), (8, 'different'), (17, 'sized'), (22, 'chunks')]

    Q: What if I want the index of a data item to be a function of the data item itself?

    A: Then you would use that function to make the ``(idxof(data_item), data_item)``
    pairs directly. ``DynamicIndexer`` is for the use case where the index of an item
    depends on the (number of, sizes of, etc.) items that came before it.

    """

    start: Index = 0
    idx_updater: IndexUpdater = count_increments

    count_increments = staticmethod(count_increments)
    size_increments = staticmethod(size_increments)

    def __post_init__(self):
        self.current_idx = self.start

    def __call__(self, x):
        _current_idx = self.current_idx
        self.current_idx = self.idx_updater(_current_idx, x)
        return _current_idx, x


def dynamically_index(iterable: Iterable, start=0, idx_updater=count_increments):
    """Generalization of `enumerate(iterable)` that allows one to specify how the
    indices should be updated.

    The default is the sae behavior as `enumerate`: Starts with 0 and increments by 1.

    >>> stream = ['stream', 'of', 'different', 'sized', 'chunks']
    >>> assert (list(dynamically_index(stream, start=2))
    ...     == list(enumerate(stream, start=2))
    ...     == [(2, 'stream'), (3, 'of'), (4, 'different'), (5, 'sized'), (6, 'chunks')]
    ... )

    Say we wanted to increment the indices according to the size of the last item
    instead of just incrementing by 1 at every iteration tick...

    >>> def size_increments(current_idx, data_item, size_func=len):
    ...     return current_idx + size_func(data_item)
    >>> size_index = DynamicIndexer(idx_updater=DynamicIndexer.size_increments)
    >>> list(map(size_index, stream))
    [(0, 'stream'), (6, 'of'), (8, 'different'), (17, 'sized'), (22, 'chunks')]

    """
    dynamic_indexer = DynamicIndexer(start, idx_updater)
    return map(dynamic_indexer, iterable)


# Alternative to the above implementation:

from itertools import accumulate


def _dynamic_indexer(stream, idx_updater: IndexUpdater = count_increments, start=0):
    index_func = partial(accumulate, func=idx_updater, initial=start)
    obj = zip(index_func(stream), stream)
    return obj


def alt_dynamically_index(idx_updater: IndexUpdater = count_increments, start=0):
    """Alternative to dynamically_index using itertools and partial

    >>> def size_increments(current_idx, data_item, size_func=len):
    ...     return current_idx + size_func(data_item)
    ...
    >>> stream = ['stream', 'of', 'different', 'sized', 'chunks']
    >>> indexer = alt_dynamically_index(size_increments)
    >>> t = list(indexer(stream))
    >>> assert t == [(0, 'stream'), (6, 'of'), (8, 'different'), (17, 'sized'),
    ...              (22, 'chunks')]
    """
    return partial(_dynamic_indexer, idx_updater=idx_updater, start=start)


# ---------------------------------------------------------------------------------------
# Slicing index segment streams


def segment_overlaps(bt_tt_segment, query_bt, query_tt):
    """Returns True if, and only if, bt_tt_segment overlaps query interval.

    A `bt_tt_segment` will need to be of the ``(bt, tt, *data)`` format.
    That is, an iterable of at least two elements (the ``bt`` and ``tt``) followed with
    more elements (the actual segment data).

    This function is made to be curried, as shown in the following example:

    >>> from functools import partial
    >>> overlapping_segments_filt = partial(segment_overlaps, query_bt=4, query_tt=8)
    >>>
    >>> list(filter(overlapping_segments_filt, [
    ...     (1, 3, 'completely before'),
    ...     (2, 4, 'still completely before (upper bounds are strict)'),
    ...     (3, 6, 'partially before, but overlaps bottom'),
    ...     (4, 5, 'totally', 'inside'),  # <- note this tuple has 4 elements
    ...     (5, 8),  # <- note this tuple has only the minimum (2) elements,
    ...     (7, 10, 'partially after, but overlaps top'),
    ...     (8, 11, 'completely after (strict upper bound)'),
    ...     (100, 101, 'completely after (obviously)')
    ... ]))  # doctest: +NORMALIZE_WHITESPACE
    [(3, 6, 'partially before, but overlaps bottom'),
    (4, 5, 'totally', 'inside'),
    (5, 8),
    (7, 10, 'partially after, but overlaps top')]

    """
    bt, tt, *segment = bt_tt_segment
    return (
        query_bt < tt <= query_tt  # the top part of segment intersects
        or query_bt <= bt < query_tt  # the bottom part of the segment intersects
        # If it's both, the interval is entirely inside the query
    )


Stats = Any
_no_value_specified_sentinel = cast(int, object())


class BufferStats(deque):
    """A callable (fifo) buffer. Calls add input to it, but also returns some results
    computed from it's contents.

    What "add" means is configurable (through ``add_new_val`` arg). Default
    is append, but can be extend etc.

    >>> bs = BufferStats(maxlen=4, func=sum)
    >>> list(map(bs, range(7)))
    [0, 1, 3, 6, 10, 14, 18]

    See what happens when you feed the same sequence again:

    >>> list(map(bs, range(7)))
    [15, 12, 9, 6, 10, 14, 18]

    More examples:

    >>> list(map(BufferStats(maxlen=4, func=''.join), 'abcdefgh'))
    ['a', 'ab', 'abc', 'abcd', 'bcde', 'cdef', 'defg', 'efgh']

    >>> from math import prod
    >>> list(map(BufferStats(maxlen=4, func=prod), range(7)))
    [0, 0, 0, 0, 24, 120, 360]

    With a different ``add_new_val`` choice.

    >>> bs = BufferStats(maxlen=4, func=''.join, add_new_val=deque.appendleft)
    >>> list(map(bs, 'abcdefgh'))
    ['a', 'ba', 'cba', 'dcba', 'edcb', 'fedc', 'gfed', 'hgfe']

    With ``add_new_val=deque.extend``, data can be fed in chunks.
    In the following, also see how we use iterize to get a function that
    takes an iterator and returns an iterator

    >>> from creek.util import iterize
    >>> window_stats = iterize(BufferStats(
    ...     maxlen=4, func=''.join, add_new_val=deque.extend)
    ... )
    >>> chks = ['a', 'bc', 'def', 'gh']
    >>> for x in window_stats(chks):
    ...     print(x)
    a
    abc
    cdef
    efgh

    Note: To those who might think that they can optimize this for special
    cases: Yes you can.
    But SHOULD you? Is it worth the increase in complexity and reduction in
    flexibility?
    See https://github.com/thorwhalen/umpyre/blob/master/misc/performance_of_rolling_window_stats.md

    """

    # __name__ = 'BufferStats'

    def __init__(
        self,
        values=(),
        maxlen: int = _no_value_specified_sentinel,
        func: Callable = sum,
        add_new_val: Callable = deque.append,
    ):
        """

        :param maxlen: Size of the buffer
        :param func: The function to be computed (on buffer contents) and
        returned when buffer is "called"
        :param add_new_val: The function that adds values on the buffer.
        Signature must be (self, new_val)
            Is usually a deque method (``deque.append`` by default, but could
            be ``deque.extend``, ``deque.appendleft`` etc.).
            Can also be any other function that
            has a valid (self, new_val) signature.
        """
        if maxlen is _no_value_specified_sentinel:
            raise TypeError('You are required to specify maxlen')
        if not isinstance(maxlen, int):
            raise TypeError(f'maxlen must be an integer, was: {maxlen}')

        super().__init__(values, maxlen=maxlen)
        self.func = func
        if isinstance(add_new_val, str):
            # assume add_new_val is a method of deque:
            add_new_val = getattr(self, add_new_val)
        self.add_new_val = add_new_val
        self.__name__ = 'BufferStats'

    def __call__(self, new_val) -> Stats:
        self.add_new_val(self, new_val)  # add the new value
        return self.func(self)


def is_not_none(x):
    return x is not None


def return_buffer_on_stats_condition(
    stats: Stats, buffer: Iterable, cond: Callable = is_not_none, else_val=None
):
    """

    >>> return_buffer_on_stats_condition(
    ... stats=3, buffer=[1,2,3,4], cond=lambda x: x%2 == 1
    ... )
    [1, 2, 3, 4]
    >>> return_buffer_on_stats_condition(
    ... stats=3, buffer=[1,2,3,4], cond=lambda x: x%2 == 0, else_val='3 is not even!'
    ... )
    '3 is not even!'
    """

    if cond(stats):
        return buffer
    else:
        return else_val


@dataclass
class Segmenter:
    """

    >>> gen = iter(range(200))
    >>> bs = BufferStats(maxlen=10, func=sum)
    >>> return_if_stats_is_odd = partial(
    ...     return_buffer_on_stats_condition,
    ...     cond=lambda x: x%2 == 1, else_val='The sum is not odd!'
    ... )
    >>> seg = Segmenter(buffer=bs, stats_buffer_callback=return_if_stats_is_odd)

    Since the sum of the values in the buffer [1] is odd, the buffer is returned:

    >>> seg(new_val=1)
    [1]

    Adding ``1 + 2`` is still odd so:

    >>> seg(new_val=2)
    [1, 2]

    Now since ``1 + 2 + 5`` is even, the ``else_val`` of ``return_if_stats_is_odd``
    is returned instead

    >>> seg(new_val=5)
    'The sum is not odd!'
    """

    buffer: BufferStats
    stats_buffer_callback: Callable[
        [Stats, Iterable], Any
    ] = return_buffer_on_stats_condition
    __name__ = 'Segmenter'

    def __call__(self, new_val):
        stats = self.buffer(new_val)
        return self.stats_buffer_callback(stats, list(self.buffer))



"""Utils to deal with async iteration


Making singledispatch work:

.. code-block:: python

    from typing import Generator, Iterator, Iterable, AsyncIterable, AsyncIterator
    from functools import singledispatch, partial

    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class IterableType(Protocol):
        def __iter__(self):
            pass

    @runtime_checkable
    class CursorFunc(Protocol):
        def __call__(self):
            pass

    @singledispatch
    def to_iterator(x: IterableType):
        return iter(x)

    will_never_happen = object()

    @to_iterator.register
    def _(x: CursorFunc):
        return iter(x, will_never_happen)

    assert list(to_iterator([1,2,3])) == [1, 2, 3]

    f = partial(next, iter([1,2,3]))
    assert list(to_iterator(f)) == [1, 2, 3]

Trying to make async iterators/iterables/cursor_funcs utils

.. code-block:: python

    import asyncio


    async def ticker(to=3, delay=0.5):
        # Yield numbers from 0 to `to` every `delay` seconds.
        for i in range(to):
            yield i
            await asyncio.sleep(delay)

    async def my_aiter(async_iterable):
        async for i in async_iterable:
            yield i

    t = [i async for i in my_aiter(ticker(3, 0.2))]
    assert t == [0, 1, 2]

    # t = list(my_aiter(ticker(3, 0.2)))
    # # TypeError: 'async_generator' object is not iterable
    # # and
    # t = await list(ticker(3, 0.2))
    # # TypeError: 'async_generator' object is not iterable

    # But...

    async def alist(async_iterable):
        return [i async for i in async_iterable]

    t = await alist(ticker(3, 0.2))
    assert t == [0, 1, 2]


"""

from functools import partial
from typing import (
    Callable,
    Any,
    NewType,
    Iterable,
    AsyncIterable,
    Iterator,
    AsyncIterator,
    Union,
)

IterableType = Union[Iterable, AsyncIterable]
IteratorType = Union[Iterator, AsyncIterator]
CursorFunc = NewType('CursorFunc', Callable[[], Any])
CursorFunc.__doc__ = "An argument-less function returning an iterator's values"


# ---------------------------------------------------------------------------------------
# iteratable, iterator, cursors
no_sentinel = type('no_sentinel', (), {})()

try:
    aiter  # exists in python 3.10+
    # Note: doesn't have the sentinel though!!
except NameError:

    async def aiter(iterable: AsyncIterable) -> AsyncIterator:
        if not isinstance(iterable, AsyncIterable):
            raise TypeError(f'aiter expected an AsyncIterable, got {type(iterable)}')
        if isinstance(iterable, AsyncIterator):
            return iterable
        return (i async for i in iterable)


async def aiter_with_sentinel(cursor_func: CursorFunc, sentinel: Any) -> AsyncIterator:
    """Like iter(async_callable, sentinel) builtin but for async callables"""
    while (value := await cursor_func()) is not sentinel:
        yield value


def iterable_to_iterator(iterable: IterableType) -> IteratorType:
    """Get an iterator from an iterable (whether async or not)

    >>> iterable = [1, 2, 3]
    >>> iterator = iterable_to_iterator(iterable)
    >>> assert isinstance(iterator, Iterator)
    >>> assert list(iterator) == iterable
    """
    if isinstance(iterable, AsyncIterable):
        return aiter(iterable)
    return iter(iterable)


def iterator_to_cursor(iterator: Iterator) -> CursorFunc:
    return partial(next, iterator)


def cursor_to_iterator(cursor: CursorFunc, sentinel=no_sentinel) -> Iterator:
    return iter(cursor, no_sentinel)


def iterable_to_cursor(iterable: Iterable, sentinel=no_sentinel) -> CursorFunc:
    iterator = iterable_to_iterator(iterable)
    if sentinel is no_sentinel:
        return iterator_to_cursor(iterator)
    else:
        return partial(next, iterator, sentinel)



import pytest
from creek.labeling import (
    DictLabeledElement,
    SetLabeledElement,
    ListLabeledElement,
)


def test_dummy():
    assert True


def test_dict_labeledelement():
    x = DictLabeledElement(42).add_label({'string': 'forty-two'})
    assert x.element == 42
    assert x.labels == {'string': 'forty-two'}
    x.add_label({'type': 'number', 'prime': False})
    assert x.element == 42
    assert x.labels == {
        'string': 'forty-two',
        'type': 'number',
        'prime': False,
    }


def test_set_labeledelement():
    x = SetLabeledElement(42).add_label('forty-two')
    assert x.element == 42
    assert x.labels == {'forty-two'}
    x.add_label('number')
    assert x.element == 42
    assert x.labels == {'forty-two', 'number'}


def test_list_labeledelement():
    x = ListLabeledElement(42).add_label('forty-two')
    assert x.element == 42
    assert x.labels == ['forty-two']
    x.add_label('number')
    assert x.element == 42
    assert x.labels == ['forty-two', 'number']
